import { Grammar } from "../grammar.js";
import { TMRUtility } from "../tmr-utility.js";
import { tmrConstants, tmrTokenZIndex } from "../tmr-constants.js";
import { Draconique } from "./draconique.js";

export class ReserveExtensible extends Draconique {
  constructor() {
    super();
  }

  type() { return 'tete' }
  match(item) { return Draconique.isTeteDragon(item) && Grammar.toLowerCaseNoAccent(item.name).includes("reserve extensible"); }
  manualMessage() { return "Vous pouvez re-configurer votre Réserve extensible" }
  async onActorCreateOwned(actor, tete) {
    const existants = actor.items.filter(it => this.isCase(it)).map(it => it.system.coord);
    const selectedTMR = await TMRUtility.getTMRAleatoire(tmr => !(tmr.type == 'fleuve' || existants.includes(tmr.coord)));
    await this.createCaseTmr(actor, "Nouvelle Réserve extensible", selectedTMR, tete.id);
  }

  code() { return 'reserve_extensible' }
  tooltip(linkData) { return `Réserve extensible en ${this.tmrLabel(linkData)} !` }
  img() { return 'systems/foundryvtt-reve-de-dragon/icons/tmr/treasure-chest.webp' }

  createSprite(pixiTMR) {
    return pixiTMR.sprite(this.code(), {
      zIndex: tmrTokenZIndex.tetes,
      alpha: 0.7,
      decallage: tmrConstants.left
    });
  }

}
